/**************************************************************************************************

Constraints.C -- (C) Niklas Een, Niklas Srensson, 2004

Defines the constraints supported by MiniSat (clauses being the core constraint). 

**************************************************************************************************/

#include "Constraints.h"
#include "Solver.h"
#include "Sort.h"
#include "SolverTypes.h"


//=================================================================================================
// Helpers:


void removeWatch(vec<Constr*>& ws, Constr* elem)
{
    int j = 0;
    for (; ws[j] != elem  ; j++) assert(j < ws.size());
    for (; j < ws.size()-1; j++) ws[j] = ws[j+1];
    ws.pop();
}


//=================================================================================================
// Clause constraint:


// Returns FALSE if top-level conflict detected (must be handled); TRUE otherwise.
// 'out_clause' may be set to NULL if clause is already satisfied by the top-level assignment.
//
bool Clause_new(Solver& S, const vec<Lit>& ps, bool learnt, Clause*& out_clause)
{
	if (&out_clause != NULL) out_clause = NULL;
    
    if (ps.size() == 0)
        return false;
    else if (ps.size() == 1)
        return S.enqueue(ps[0]);
    else{
        // Allocate clause:
        assert(sizeof(Lit)   == sizeof(unsigned));
        assert(sizeof(float) == sizeof(unsigned));
        void*   mem = xmalloc<char>(sizeof(Clause) + sizeof(unsigned)*(ps.size() + 1));
        Clause* c   = new (mem) Clause;

		c->size_learnt = (int)learnt | (ps.size() << 1);
        for (int i = 0; i < ps.size(); i++) c->data[i] = ps[i];
        if (learnt) c->activity() = 0.0;

        // For learnt clauses only:
        if (learnt){
            // Put the second watch on the literal with highest decision level:
            int     max_i = 1;
            int     max   = S.level[var(ps[1])];
            for (int i = 2; i < ps.size(); i++)
                if (S.level[var(ps[i])] > max)
                    max   = S.level[var(ps[i])],
                    max_i = i;
            c->data[1]     = ps[max_i];
            c->data[max_i] = ps[1];

            // Bumping:
            S.claBumpActivity(c); // (newly learnt clauses should be considered active)
            for (int i = 0; i < ps.size(); i++)
                S.varBumpActivity(ps[i]);
        } 
        // Store clause:
        S.watches[index(~c->data[0])].push(c);
        S.watches[index(~c->data[1])].push(c);
        if (&out_clause != NULL) out_clause = c;

        return true;
    }
}
//=================================================================================================
// Builds new clause from integer input vector
               
bool Clause::constr_new(Solver& S, const vec<int>& ints, Constr*& out_clause){
    vec<Lit>    qs;
    if (&out_clause != NULL) out_clause = NULL;

    assert(S.decisionLevel() == 0);
	    
	qs.growTo(ints.size());  // Make a copy of the input vector.
	for(int i=0;i<ints.size();++i){
		int var = abs(ints[i])-1;
        while (var >= S.nVars()) S.newVar();
		new (&qs[i]) Lit(parseToLit(ints[i]));
	}

    // Remove false literals:
    for (int i = 0; i < qs.size();){
        if (S.value(qs[i]) != l_Undef){
            if (S.value(qs[i]) == l_True)
                return true;    // Clause always true -- don't add anything.
            else
                qs[i] = qs.last(),
                qs.pop();
        }else
            i++;
    }

    // Remove duplicates:
    sortUnique(qs);
    for (int i = 0; i < qs.size()-1; i++){
        if (qs[i] == ~qs[i+1])
           return true;        // Clause always true -- don't add anything.
    }
	// 'qs' is now the (possibly) reduced vector of literals.
	Clause *c;
	bool ok=Clause_new(S, qs, false, c);
	out_clause=c;
	return ok;
}

//=================================================================================================
bool Clause::locked(const Solver& S) const {
    return (const Clause *)S.reason[var(data[0])] == this; }


void Clause::remove(Solver& S, bool just_dealloc)
{
    if (!just_dealloc){
        removeWatch(S.watches[index(~data[0])], this);
        removeWatch(S.watches[index(~data[1])], this); }
    xfree(this);
}


// Can assume everything has been propagated! (esp. the first two literals are != l_False, unless
// the clause is binary and satisfied, in which case the first literal is true)
// Returns True if clause is satisfied (will be removed), False otherwise.
//
bool Clause::simplify(Solver& S)
{
    assert(S.decisionLevel() == 0);
    for (int i = 0; i < size(); i++){
        if (S.value(data[i]) == l_True)
            return true;
    }
    return false;
}


// 'p' is the literal that became TRUE
bool Clause::propagate(Solver& S, Lit p, bool& keep_watch)
{
    // Make sure the false literal is data[1]:
    Lit     false_lit = ~p;
    if (data[0] == false_lit)
        data[0] = data[1], data[1] = false_lit;
    assert(data[1] == false_lit);

    // If 0th watch is true, then clause is already satisfied.
    if (S.value(data[0]) == l_True){
        keep_watch = true;
        return true; }

    // Look for new watch:
    for (int i = 2; i < size(); i++){
        if (S.value(data[i]) != l_False){
            data[1] = data[i], data[i] = false_lit;
            S.watches[index(~data[1])].push(this);
            return true; } }

    // Clause is unit under assignment:
    keep_watch = true;
    return S.enqueue(data[0], this);
}


// Can assume 'out_reason' to be empty.
// Calculate reason for 'p'. If 'p == lit_Undef', calculate reason for conflict.
//
void Clause::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{
    assert(p == lit_Undef || p == data[0]);
    for (int i = ((p == lit_Undef) ? 0 : 1); i < size(); i++)
        assert(S.value(data[i]) == l_False),
        out_reason.push(~data[i]);
    if (learnt()) S.claBumpActivity(this);
}


//=================================================================================================
// AtMost constraint 
//
bool AtMost::constr_new(Solver& S, const vec<int>& ints, Constr*& out_clause)
{
	int n = ints[0];
	out_clause=0;
    vec<Lit>    qs;
    assert(S.decisionLevel() == 0);
	    
	qs.growTo(ints.size()-1);  // Make a copy of the input vector except of ints[0]	
	for(int i=1;i<ints.size();++i){
		int var = abs(ints[i])-1;
        while (var >= S.nVars()) S.newVar();
		new (&qs[i-1]) Lit(parseToLit(ints[i]));
	}

    // Remove true, false literals:
    for (int i = 0; i < qs.size();){
        if (S.value(qs[i]) != l_Undef){
			if (S.value(qs[i]) == l_True){
                qs[i] = qs.last(),
                qs.pop();
				--n;
			}
            else
                qs[i] = qs.last(),
                qs.pop();
        }else
            i++;
    }

    // Remove duplicates:
    sortUnique(qs);
    for (int i = 0; i < qs.size()-1; i++)
		if (qs[i] == ~qs[i+1]){
			qs[i]  = qs.last();
			qs[i+1]= qs[qs.size()-2];
            qs.pop();
			qs.pop();
			--n;
		}
               
    int size = qs.size();	
	if(size <= n)return true; //tautology
	if(n==0){
		
		for(int i=0;i<size;++i)
			if(!S.enqueue(~qs[i]))
				return false;
		return true;
	}
    void* mem    = (void*)xmalloc<char>(sizeof(AtMost) + sizeof(Lit) * qs.size());
    AtMost *out  = new (mem) AtMost;
    out->size    = size;
    out->n       = n;
    out->counter = 0;
    for (int i = 0; i < qs.size(); i++) out->lits[i] = qs[i];
    for (int i = 0; i < qs.size(); i++) S.watches[index(qs[i])].push(out);
		
	out_clause=out;
    return true;
}


void AtMost::remove(Solver& S, bool just_dealloc) {
    if (!just_dealloc)
        for (int i = 0; i < size; i++)
            removeWatch(S.watches[index(lits[i])], this);
    xfree(this);
}


bool AtMost::simplify(Solver& S) {
    return false; }


bool AtMost::propagate(Solver& S, Lit p, bool& keep_watch)
{
    keep_watch = true;
    if (counter == n) return false;

    counter++;
    S.undos[var(p)].push(this);

    // If no more can be true, enqueue the negation of the rest:
    if (counter == n)
        for (int i = 0; i < size; i++)
            if ((S.value(lits[i]) == l_Undef) && !S.enqueue(~lits[i], this))
                return false;
    return true;
}


void AtMost::undo(Solver& S, Lit p) {
    counter--; }


void AtMost::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{
    int     c = (p == lit_Undef) ? -1 : 0;
    for (int i = 0; i < size; i++){
        if (S.value(lits[i]) == l_True){
            out_reason.push(lits[i]);
			if (++c == n) return;
			
        }
    }
}


//=================================================================================================
// Xor constraint 



bool XorConstr::constr_new(Solver& S, const vec<int>& ints, Constr*& out_clause)
{
	out_clause=0;
    vec<Lit>    qs;
    assert(S.decisionLevel() == 0);
	    
	qs.growTo(ints.size());  // Make a copy of the input vector.
	for(int i=0;i<ints.size();++i){
		int var = abs(ints[i])-1;
        while (var >= S.nVars()) S.newVar();
		new (&qs[i]) Lit(parseToLit(ints[i]));
	}

    // Remove true, false literals:
    for (int i = 0; i < qs.size();){
        if (S.value(qs[i]) != l_Undef){
			if (S.value(qs[i]) == l_True){
                qs[i] = qs.last(),
                qs.pop();
				if(qs.size()>0)qs[0]= ~qs[0];//invert any varible
			}
            else
                qs[i] = qs.last(),
                qs.pop();
        }else
            i++;
    }

    // Remove duplicates:
    sortUnique(qs);
    for (int i = 0; i < qs.size()-1; i++)
		if (qs[i] == ~qs[i+1]){
			qs[i]  = qs.last();
			qs[i+1]= qs[qs.size()-2];
            qs.pop();
			qs.pop();
		}
               
    int size = qs.size();
	if(size <= 0)return true; //tautology	
	if(size==1){
		return S.enqueue(~qs[0]);		
	}
    void* mem    = (void*)xmalloc<char>(sizeof(XorConstr) + sizeof(Lit) * qs.size());
    XorConstr *out  = new (mem) XorConstr;
    out->size    = size;
    for (int i = 0; i < qs.size(); i++) out->lits[i] = qs[i];
	for (int i = 0; i < qs.size(); i++){
		S.watches[index(qs[i])].push(out);
		S.watches[index(~qs[i])].push(out);
	}
		
	out_clause=out;
    return true;
}


void XorConstr::remove(Solver& S, bool just_dealloc) {
	if (!just_dealloc){        
		removeWatch(S.watches[index(lits[0])], this);
		removeWatch(S.watches[index(lits[1])], this);
	}
    xfree(this);
}


bool XorConstr::simplify(Solver& S) {
    return false; 
}


bool XorConstr::propagate(Solver& S, Lit p, bool& keep_watch)
{
    // Make sure the new literal is data[1]:
    Lit     tmp_lit = ~p;
	bool res=false;
	if (lits[0]==p || lits[0] == ~p){
        tmp_lit=lits[0];
		lits[0] = lits[1];
		lits[1] = tmp_lit;
	}
    
    // Look for new watch:
    for (int i = 1; i < size; i++){
		if (S.value(lits[i]) == l_Undef){
			tmp_lit = lits[1];
			lits[1] = lits[i];
			lits[i] = tmp_lit;
			keep_watch = false;
			removeWatch(S.watches[index(~p)], this);
			S.watches[index(lits[1])].push(this);
			S.watches[index(~lits[1])].push(this);
			return true;
		}else if(S.value(lits[i]) == l_True){
			res= !res;
		};
	};
        
    // Clause is unit under assignment:
    keep_watch = true;
	return S.enqueue(res ? lits[0] : ~lits[0], this);    
}


void XorConstr::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{
	for (int i = ((p == lit_Undef) ? 0 : 1); i < size; i++)
		out_reason.push((S.value(lits[i])==l_True) ? lits[i] : ~lits[i]);

}

//=================================================================================================
// MatrixC constraint 
//
bool MatrixC::constr_new(Solver& S, const vec<int>& ints, Constr*& out_clause)
{	
	out_clause=0;

    void* mem    = (void*)xmalloc<char>(sizeof(MatrixC));
    MatrixC *out  = new (mem) MatrixC;
    out->size    = ints[0];
	out->rows = ((ints.size()-1)/out->size)-1;
	out->boolMatrix = (lbool*)xmalloc<lbool>(out->size * out->rows);	
	out->lits = (Lit*)xmalloc<Lit>(out->size);
	out->result =(lbool*)xmalloc<lbool>(out->size);	

	for(int i=1;i<=out->size;++i){
		int var = abs(ints[i])-1;
        while (var >= S.nVars()) S.newVar();
		new (&out->lits[i-1]) Lit(parseToLit(ints[i]));    
	}
	
	for(int i=0,p=out->size+1,q=0;i<out->rows;++i){		
		for(int col=0;col<out->size;++col,++p,++q){
			out->boolMatrix[q]=toLbool(ints[p]-2); //offset 1, 2, 3 -> false, unknown, true 
		}								  //which is -1, 0, 1
	}

	for (int i = 0; i < out->size; i++) {
		S.watches[index(out->lits[i])].push(out);
		S.watches[index(~out->lits[i])].push(out);
	}
		
	out_clause=out;
	Lit tmp;
	bool tmpb;
    return out->propagate(S,tmp,tmpb);
}


void MatrixC::remove(Solver& S, bool just_dealloc) {
    if (!just_dealloc)
		for (int i = 0; i < size; i++){
            removeWatch(S.watches[index(lits[i])], this);
			removeWatch(S.watches[index(~lits[i])], this);
		}
	xfree(this->lits);
	xfree(this->result);
	xfree(this->boolMatrix);	
    xfree(this);
}


bool MatrixC::simplify(Solver& S) {
    return false; }

inline bool conflict(lbool& a, lbool& b){
	return (a==l_True && b==l_False || a==l_False && b==l_True);
}

lbool lbooland[][4]={{toLbool(-2),toLbool(-2),toLbool(-2),toLbool(-2)},
                     {toLbool(-2),toLbool(-1),toLbool(-2),toLbool(-2)},
                     {toLbool(-2),toLbool(-1),toLbool(-2),toLbool( 1)},
                     {toLbool(-2),toLbool(-2),toLbool(-2),toLbool( 1)}};

bool MatrixC::propagate(Solver& S, Lit p, bool& keep_watch)
{
    keep_watch = true;	
    for(int i=0;i<size;++i)result[i]=l_Undef;

	for(int i=0,isize=0;i<rows;++i,isize+=size){
		bool rowOk=true;
		for(int col=0;col<size;++col){
			if(conflict(S.value(var(lits[col])), boolMatrix[isize+col])){
				rowOk=false;
				break;			
			}
		}
		if(!rowOk)continue;
		for(int col=0;col<size;++col){			
			result[col]= lbooland[toInt(result[col])+2][toInt(boolMatrix[isize+col])+2];			
		}
		
	}
	bool ok=true;
	for(int col=0;col<size;++col)
		if(result[col]==l_True)
			ok&=S.enqueue(lits[col],this);
		else if (result[col]==l_False)
			ok&=S.enqueue(~lits[col],this);    

    return ok;
}




void MatrixC::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{    
    for (int i = 0; i < size; i++){
        if (S.value(lits[i]) != l_Undef && lits[i]!=p){
            out_reason.push((S.value(lits[i])==l_True) ? lits[i] : ~lits[i]);						
        }
    }
}

