/**************************************************************************************************

Constraints.h -- (C) Niklas Een, Niklas Srensson, 2004
Extended by Antonin Lejsek, 2005

Defines the constraints supported by MiniSat (clauses being the core constraint). 

**************************************************************************************************/

#ifndef Constraints_h
#define Constraints_h

#include "SolverTypes.h"


//=================================================================================================
// Constraint abstraction:


class Solver;

struct Constr {
	virtual bool constr_new(Solver& S, const vec<int>& ints, Constr*& out_clause)=0;
    virtual void remove    (Solver& S, bool just_dealloc = false) = 0;
    virtual bool propagate (Solver& S, Lit p, bool& keep_watch) = 0;    // ('keep_watch' is set to FALSE beftore call to this method)
    virtual bool simplify  (Solver& S) { return false; };
    virtual void undo      (Solver& S, Lit p) { };
    virtual void calcReason(Solver& S, Lit p, vec<Lit>& out_reason) = 0;
	virtual const char* getType()=0;

    virtual ~Constr(void) { };  // (not used, just to keep the compiler happy)
};


//=================================================================================================
// Clauses:


class Clause : public Constr {
    unsigned    size_learnt;
    Lit         data[0];

public:
    int  size        (void)      const { return size_learnt >> 1; }
    bool learnt      (void)      const { return size_learnt & 1; }
    Lit  operator [] (int index) const { return data[index]; }

    // Constructor -- creates a new clause and add it to watcher lists. 
    friend bool Clause_new(Solver& S, const vec<Lit>& ps, bool learnt, Clause*& out_clause);
	bool constr_new(Solver& S, const vec<int>& ints, Constr*& out_clause);
	               

    // Learnt clauses only:
    bool    locked  (const Solver& S) const;
    float&  activity(void) const { return *((float*)&data[size()]); }

    // Constraint interface:
    void remove    (Solver& S, bool just_dealloc = false);
    bool propagate (Solver& S, Lit p, bool& keep_watch);
    bool simplify  (Solver& S);
    void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);
	const char* getType(){return "clause";};
};

//=================================================================================================
// AtMost:


class AtMost : public Constr {
    int     n;
    int     counter;
    int     size;
    Lit     lits[0];

public:       

    // Constraint interface:
    void remove    (Solver& S, bool just_dealloc = false);
    bool propagate (Solver& S, Lit p, bool& keep_watch);
    bool simplify  (Solver& S);
    void undo      (Solver& S, Lit p);
    void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);
	bool constr_new(Solver& S, const vec<int>& ints, Constr*& out_clause);
	const char* getType(){return "atmost";};
};



//=================================================================================================
// Xor:


class XorConstr : public Constr {
    int     size;
    Lit     lits[0];

public:       

    // Constraint interface:
    void remove    (Solver& S, bool just_dealloc = false);
    bool propagate (Solver& S, Lit p, bool& keep_watch);
    bool simplify  (Solver& S);
	void undo      (Solver& S, Lit p){;};
    void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);
	bool constr_new(Solver& S, const vec<int>& ints, Constr*& out_clause);
	const char* getType(){return "xor";};
};



//=================================================================================================

// MatrixC:


class MatrixC : public Constr {
    int     size;    //number of variables
	int		rows;    //number of matrix rows
    Lit*    lits;    //literals
	lbool*    result;  //temporary space for result	
	lbool* boolMatrix; //Matrix of all acceptable boolean combinations size*rows


public:       

    // Constraint interface:
    void remove    (Solver& S, bool just_dealloc = false);
    bool propagate (Solver& S, Lit p, bool& keep_watch);
    bool simplify  (Solver& S);
	void undo      (Solver& S, Lit p){;};
    void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);
	bool constr_new(Solver& S, const vec<int>& ints, Constr*& out_clause);
	const char* getType(){return "matrix";};
};



//=================================================================================================



#endif
